// pcm.c

#include "amy.h"



#define PCM_AMY_LOG2_SAMPLE_RATE log2f(PCM_AMY_SAMPLE_RATE / ZERO_LOGFREQ_IN_HZ)

void pcm_init() {

}

// How many bits used for fractional part of PCM table index.
#define PCM_INDEX_FRAC_BITS 8
// The number of bits used to hold the table index.
#define PCM_INDEX_BITS (31 - PCM_INDEX_FRAC_BITS)

void pcm_note_on(uint16_t osc) {
    //printf("pcm_note_on: osc=%d patch=%d logfreq=%f amp=%f\n",
    //       osc, synth[osc].patch, synth[osc].logfreq, synth[osc].amp);
    if(synth[osc].patch >= pcm_samples) synth[osc].patch = 0;
    // if no freq given, just play it at midinote
    if(synth[osc].logfreq_coefs[COEF_CONST] <= 0) {
        // This will result in PCM_SAMPLE_RATE when the midi_note == patch->midinote.
        synth[osc].logfreq_coefs[COEF_CONST] = PCM_AMY_LOG2_SAMPLE_RATE - logfreq_for_midi_note(pcm_map[synth[osc].patch].midinote);
    }
    synth[osc].phase = 0; // s16.15 index into the table; as if a PHASOR into a 16 bit sample table.
    // Special case: We use the msynth feedback flag to indicate note-off for looping PCM.  As a result, it's explicitly NOT set in amy:hold_and_modify for PCM voices.  Set it here.
    msynth[osc].feedback = synth[osc].feedback;
}

void pcm_mod_trigger(uint16_t osc) {
    pcm_note_on(osc);
}

void pcm_note_off(uint16_t osc) {
    if(msynth[osc].feedback == 0) {
        // Non-looping note: Set phase to the end to cause immediate stop.
        synth[osc].phase = F2P(pcm_map[synth[osc].patch].length / (float)(1 << PCM_INDEX_FRAC_BITS));
    } else {
        // Looping is requested, disable future looping, sample will play through to end.
        // (sending a second note-off will stop it immediately).
        msynth[osc].feedback = 0;
    }
}

SAMPLE render_pcm(SAMPLE* buf, uint16_t osc) {
    // Patches can be > 32768 samples long.
    // We need s16.15 fixed-point indexing.
    const pcm_map_t* patch = &pcm_map[synth[osc].patch];
    float logfreq = msynth[osc].logfreq;
    // If osc[midi_note] is unset, apply patch's default here.
    if (AMY_IS_UNSET(synth[osc].midi_note))  logfreq += logfreq_for_midi_note(patch->midinote);
    float playback_freq = freq_of_logfreq(logfreq);  // PCM_SAMPLE_RATE modified by

    SAMPLE max_value = 0;
    SAMPLE amp = F2S(msynth[osc].amp);
    PHASOR step = F2P((playback_freq / (float)AMY_SAMPLE_RATE) / (float)(1 << PCM_INDEX_BITS));
    const LUTSAMPLE* table = pcm + patch->offset;
    uint32_t base_index = INT_OF_P(synth[osc].phase, PCM_INDEX_BITS);
    //fprintf(stderr, "render_pcm: time=%.3f patch=%d base_index=%d length=%d loopstart=%d loopend=%d fb=%f is_unset_note_off %d\n", total_samples / (float)AMY_SAMPLE_RATE, synth[osc].patch, base_index, patch->length, patch->loopstart, patch->loopend, msynth[osc].feedback, AMY_IS_UNSET(synth[osc].note_off_clock));
    for(uint16_t i=0; i < AMY_BLOCK_SIZE; i++) {
        SAMPLE frac = S_FRAC_OF_P(synth[osc].phase, PCM_INDEX_BITS);
        LUTSAMPLE b = table[base_index];
        LUTSAMPLE c = b;
        if (base_index < patch->length) c = table[base_index + 1];
        SAMPLE sample = L2S(b) + MUL0_SS(L2S(c - b), frac);
        synth[osc].phase = P_WRAPPED_SUM(synth[osc].phase, step);
        base_index = INT_OF_P(synth[osc].phase, PCM_INDEX_BITS);
        if(base_index >= patch->length) { // end
            synth[osc].status = STATUS_OFF;// is this right? 
            sample = 0;
        } else {
            if(msynth[osc].feedback > 0) { // still looping.  The feedback flag is cleared by pcm_note_off.
                if(base_index >= patch->loopend) { // loopend
                    // back to loopstart
                    int32_t loop_len = patch->loopend - patch->loopstart;
                    synth[osc].phase -= F2P(loop_len / (float)(1 << PCM_INDEX_BITS));
                    base_index -= loop_len;
                }
            }
        }
        SAMPLE value = buf[i] + MUL4_SS(amp, sample);
        buf[i] = value;
        if (value < 0) value = -value;
        if (value > max_value) max_value = value;        
    }
    //printf("render_pcm: osc %d patch %d len %d base_ix %d phase %f step %f tablestep %f amp %f\n",
    //       osc, synth[osc].patch, patch->length, base_index, P2F(synth[osc].phase), P2F(step), (1 << PCM_INDEX_BITS) * P2F(step), S2F(msynth[osc].amp));
    return max_value;
 }

SAMPLE compute_mod_pcm(uint16_t osc) {
    float mod_sr = (float)AMY_SAMPLE_RATE / (float)AMY_BLOCK_SIZE;
    PHASOR step = F2P((PCM_AMY_SAMPLE_RATE / mod_sr) / (1 << PCM_INDEX_BITS));
    const pcm_map_t* patch = &pcm_map[synth[osc].patch];
    const LUTSAMPLE* table = pcm + patch->offset;
    uint32_t base_index = INT_OF_P(synth[osc].phase, PCM_INDEX_BITS);
    SAMPLE sample;
    if(base_index >= patch->length) { // end
        synth[osc].status = STATUS_OFF;// is this right? 
        sample = 0;
    } else {
        sample = L2S(table[base_index]);
        synth[osc].phase = P_WRAPPED_SUM(synth[osc].phase, step);
    }
    return MUL4_SS(F2S(msynth[osc].amp), sample);
}
